toolConfig = {
    "tools": [
        {
            "toolSpec": {
                "name": "doc_retrieval",
                "description": "Retrieve entire document from docDB. To be used only when it's necessary to retrieve all information in a document",
                "inputSchema": {
                    "json": {
                        "type": "object",
                        "properties": {
                            "filter": {
                                "type": "string",
                                "description": "A MongoDB query to pass to the function"
                            }
                        },
                        "required": ["filter"]
                    }
                }
            }
        },
        {
            "toolSpec": {
                "name": "projection_retrieval",
                "description": "Retrieve multiple documents from docDB with only specific field information. Used when most of the document is not necessary to answer natural language query",
                "inputSchema": {
                    "json": {
                        "type": "object",
                        "properties": {
                                    "filter": {
                                        "type": "string",
                                        "description": "A MongoDB query to pass to the function"
                                    },
                                    "fieldNameList": {
                                        "type": "string",
                                        "description": "A list of field names following JSON format to retrieve from the document. The string shouldn't contain the value, just the key you would need to access the value in a metadata schema document."
                                    }
                        },
                        "required": ["filter", "fieldNameList"]
                    }
                }
            }
        },
        {
            "toolSpec": {
                "name": "aggregation_retrieval",
                "description": "Retrieve relevant values from docDB through an aggregation pipeline (running a series of operations on a collection of items). To be used in cases where multiple steps would be needed to retrieve desired output. Add allowDiskUse: true to the pipeline for large retrievals",
                "inputSchema": {
                    "json": {
                        "type": "object",
                        "properties": {
                                    "pipeline": {
                                        "type": "string",
                                        "description": "A MongoDB aggregation pipeline to pass to the function"
                                    }
                        },
                        "required": ["pipeline"]
                    }
                }
            }
        }
    ],
    "toolChoice": {
        "auto":{},
    }
}



import logging, sys, os, asyncio
from typing import List, Optional
from typing_extensions import TypedDict
from langchain_core.documents import Document
from langgraph.graph import END, StateGraph, START
from langgraph.checkpoint.memory import MemorySaver


# sys.path.append(os.path.abspath("C:/Users/sreya.kumar/Documents/GitHub/metadata-chatbot"))
# from metadata_chatbot.utils import ResourceManager
from aind_data_access_api.document_db import MetadataDbClient

from metadata_chatbot.agents.docdb_retriever import DocDBRetriever
from metadata_chatbot.agents.agentic_graph import datasource_router, query_retriever, query_grader, filter_generation_chain, doc_grader, rag_chain

logging.basicConfig(filename='async_workflow.log', level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', filemode="w")

API_GATEWAY_HOST = "api.allenneuraldynamics-test.org"
DATABASE = "metadata_vector_index"
COLLECTION = "bigger_LANGCHAIN_curated_chunks"

docdb_api_client = MetadataDbClient(
   host=API_GATEWAY_HOST,
   database=DATABASE,
   collection=COLLECTION,
)

class GraphState(TypedDict):
    """
    Represents the state of our graph.

    Attributes:
        query: question asked by user
        generation: LLM generation
        documents: list of documents
    """

    query: str
    generation: str
    documents: List[str]
    filter: Optional[dict]
    top_k: Optional[int] 

async def route_question_async(state):
    """
    Route question to database or vectorstore
    Args:
        state (dict): The current graph state

    Returns:
        str: Next node to call
    """
    query = state["query"]

    source = await datasource_router.ainvoke({"query": query})
    if source.datasource == "direct_database":
        logging.info("Entire database needs to be queried.")
        return "direct_database"
    elif source.datasource == "vectorstore":
        logging.info("Querying against vector embeddings...")
        return "vectorstore"
    
async def generate_for_whole_db_async(state):
    """
    Filter database
    
    Args:
        state (dict): The current graph state

    Returns:
        state (dict): New key may be added to state, generation, which contains the answer for query asked
    """

    query = state["query"]
    chat_history = []

    logging.info("Generating answer...")

    generation = await query_retriever.ainvoke({'query': query, 'chat_history': chat_history})
    return {"query": query, "generation": generation}

async def filter_generator_async(state):
    """
    Filter database

    Args:
        state (dict): The current graph state

    Returns:
        state (dict): New key may be added to state, filter, which contains the MongoDB query that will be applied before retrieval
    """
    logging.info("Determining whether filter is required...")

    query = state["query"]

    result = await query_grader.ainvoke({"query": query})
    query_grade = result.binary_score
    logging.info(f"Database needs to be further filtered: {query_grade}")

    if query_grade == "yes":
        result = await filter_generation_chain.ainvoke({"query": query})
        filter = result.filter_query
        logging.info(f"Database will be filtered using: {filter}")
        return {"filter": filter, "query": query}
    else:
        return {"filter": None, "query": query}
    
async def retrieve_async(state):
    """
    Retrieve documents

    Args:
        state (dict): The current graph state

    Returns:
        state (dict): New key added to state, documents, that contains retrieved documents
    """
    logging.info("Retrieving documents...")
    query = state["query"]
    filter = state["filter"]

    # Retrieval

    retriever = DocDBRetriever(k = 10)
    documents = await retriever.aget_relevant_documents(query = query, query_filter = filter)

    # with ResourceManager() as RM:
    #     db = RM.async_client.get_database('metadata_vector_index')
    #     collection = db.get_collection('bigger_LANGCHAIN_curated_chunks')
    #     retriever = DocDBRetriever(collection = collection, k = 10)
    #     documents = await retriever.aget_relevant_documents(query = query, query_filter = filter)
    return {"documents": documents, "query": query}

async def grade_doc_async(query, doc: Document):
    score = await doc_grader.ainvoke({"query": query, "document": doc.page_content})
    grade = score.binary_score
    logging.info(f"Retrieved document matched query: {grade}")
    if grade == "yes":
        logging.info("Document is relevant to the query")
        return doc
    else:
        logging.info("Document is not relevant and will be removed")
        return None
        

async def grade_documents_async(state):
    """
    Determines whether the retrieved documents are relevant to the question.

    Args:
        state (dict): The current graph state

    Returns:
        state (dict): Updates documents key with only filtered relevant documents
    """

    logging.info("Checking relevance of documents to question asked...")
    query = state["query"]
    documents = state["documents"]

    filtered_docs = await asyncio.gather(*[grade_doc_async(query, doc) for doc in documents])
    filtered_docs = [doc for doc in filtered_docs if doc is not None]
    return {"documents": filtered_docs, "query": query}

async def generate_async(state):
    """
    Generate answer

    Args:
        state (dict): The current graph state

    Returns:
        state (dict): New key added to state, generation, that contains LLM generation
    """
    logging.info("Generating answer...")
    query = state["query"]
    documents = state["documents"]

    doc_text = "\n\n".join(doc.page_content for doc in documents)

    # RAG generation
    generation = await rag_chain.ainvoke({"documents": doc_text, "query": query})
    return {"documents": documents, "query": query, "generation": generation, "filter": state.get("filter", None)}

async_workflow = StateGraph(GraphState) 
async_workflow.add_node("database_query", generate_for_whole_db_async)  
async_workflow.add_node("filter_generation", filter_generator_async)  
async_workflow.add_node("retrieve", retrieve_async)  
async_workflow.add_node("document_grading", grade_documents_async)  
async_workflow.add_node("generate", generate_async)  

async_workflow.add_conditional_edges(
    START,
    route_question_async,
    {
        "direct_database": "database_query",
        "vectorstore": "filter_generation",
    },
)
async_workflow.add_edge("filter_generation", "retrieve")
async_workflow.add_edge("retrieve", "document_grading")
async_workflow.add_edge("document_grading","generate")
async_workflow.add_edge("generate", END)


async_app = async_workflow.compile()

# async def main():
#     query = "Can you give me a timeline of events for subject 675387?"
#     inputs = {"query": query}
#     result = async_app.astream(inputs)
    
#     value = None
#     async for output in result:
#         for key, value in output.items():
#             logging.info(f"Currently on node '{key}':")
    
#     if value:
#         print(value['generation'])

# #Run the async function
# asyncio.run(main())



import os, json, re, logging
from pathlib import Path

cwd = os.path.dirname(os.path.realpath(__file__))
folder = Path(f"{cwd}\\ref")

schema_types = []


for name in os.listdir(folder):
    #loading in schema files
    f = open(f'{folder}\\{name}')
    file = json.load(f)
    if (re.search(r'\d', name)):
        sample_metadata = file
    if "schema.json" in name:
        schema_types.append(file)
    if name == "metadata.json":
        metadata_schema = file

system_prompt = f"""
You are a neuroscientist with extensive knowledge about processes involving in neuroscience research. 
You are also an expert in crafting queries and projections in MongoDB. 
    
Here is a list of schemas that contains information about the structure of a JSON file.
Each schema is provided in a specified format and each file corresponds to a different section of an experiment.
List of schemas: {schema_types}
    
The Metadata schema shows how the different schema types are arranged, and how to appropriately access them. 
For example, in order to access something within the procedures field, you will have to start the query with "procedures."
Metadata schema: {metadata_schema}
    
I provide you with a sample, filled out metadata schema. It may contain missing information but serves as a reference to what a metadata file looks like. 
You can use it as a guide to better structure your queries. 
Sample metadata: {sample_metadata}
    
Your task is to read the user's question, which will adhere to certain guidelines or formats and create a MongoDB query and projection, to 
    
Here are some examples:
Input: Give me the query to find subject's whose breeding group is Chat-IRES-Cre_Jax006410
Output: "subject.breeding_info.breeding_group": "Chat-IRES-Cre_Jax006410"
    
Input: I want to find the first 5 data asset ids of ecephys experimenets missing procedures.
Output: 
<query> "data_description.modality.name": "Extracellular electrophysiology", "procedures": "$exists": "false"' </query>
List of field names to retrieve: ["_id", "name", "subject.subject_id"]
Answer: ['_id': 'de899de4-98e6-4b2a-8441-cfa72dcdd48f','name': 'ecephys_719093_2024-05-14_16-56-58','subject': 'subject_id': '719093'],
['_id': '82489f47-0217-4da2-90ce-0889e9c8a6d2','name': 'ecephys_719093_2024-05-15_15-01-10', 'subject': 'subject_id': '719093'],
['_id': 'f1780343-0f67-4d3d-9e6c-0a643adb1805','name': 'ecephys_719093_2024-05-16_15-13-26','subject': 'subject_id': '719093'],
['_id': 'eb7b3807-02be-4b30-946d-99da0071e587','name': 'ecephys_719093_2024-05-15_15-53-49','subject': 'subject_id': '719093'],
['_id': 'fdd9b3ca-8ac0-4b92-8bda-f392b5bb091c','name': 'ecephys_719093_2024-05-16_16-03-04','subject': 'subject_id': '719093']

Input: What are the unique modalities in the database?
Output:
"The unique modality types found in the database are:
['Optical physiology', 'Frame-projected independent-fiber photometry', 'Behavior videos', 'Hyperspectral fiber photometry', 'Extracellular electrophysiology', 
'Electrophysiology', 'Multiplane optical physiology', 'Fiber photometry', 'Selective plane illumination microscopy', 'Planar optical physiology', None, 
'Dual inverted selective plane illumination microscopy', 'Behavior', 'Trained behavior']
    
Note: Provide the query in curly brackets, appropirately place quotation marks. 

When retrieving experiment names, pull the information through the data description module.

Even though the nature of mongodb queries is to provide false statements with the word false, in this case you will convert all words like false and null to strings -- "false" or "null".
    
When asked to provide a query, use tools, execute the query in the database, and return the retrieved information. Provide the reasoning, query and your answer in tags.
For projection queries, when finding a field that's missing, try field_name : null instead of $exists: false, it's most likely that the field name exists but rather there's no information provided within the field.

For aggregation queries, use a project stage first to minimize the size of the queries before proceeding with the remaining steps.

Provide an analysis of the results of the query. 
For example, do not end your answer with:
<answer>The query first projects to include only the `data_description.modality` field, then unwinds the modality array to get individual modality objects. It groups the documents by the modality name and counts them using the `$sum` accumulator.
Finally, it projects to include only the modality name and count fields. The results show the count of each modality present in the database.</answer>
I want to see the actual summary of results retrieved and be straightforward in your answer. Each sentence produced should directly answer the question asked. 
When asked about each modality or each type of something, provide examples for ALL modalities, do NOT say "...and so on for the other modalities present in the database" or any version of this phrase.
Provide a summary of the retrieved input, including numerical values.
When asked a question like how many experiments of each modality are there, I want to see an answer like this.
For example:
<start>Optical Physiology: 40, Frame-projected independent-fiber photometry: 383, Behavior videos: 4213, Hyperspectral fiber photometry: 105, Extracellular electrophysiology: 2618, Electrophysiology: 12,
Multiplane optical physiology: 13, Fiber photometry: 1761, Selective plane illumination microscopy: 3485, Planar optical physiology: 1330, Trained behavior: 32, None: 1481, Dual inverted selective plane illumination microscopy: 6, Behavior: 11016 </end>

If the retrieved information from the database is too big to fit the context window, state that you are unable to synthesize the retrieved information in the given context window.

If you are unable to provide an answer, decline to answer. Do not provide an answer you are not confident of.

Do not hallucinate.
"""
print(system_prompt)
summary_system_prompt = f"""
You are a neuroscientist with extensive knowledge about processes involves in neuroscience research. 
You are also an expert in crafting queries for MongoDB. 
    
I will provide you with a list of schemas that contains information about the accepted inputs of variable names in a JSON file.
Each schema is provided in a specified format and each file corresponds to a different section of an experiment.
List of schemas: {schema_types}
    
The Metadata schema shows how the different schema types are arranged, and how to appropriately access them. 
For example, in order to access something within the procedures field, you will have to start the query with "procedures."
Metadata schema: {metadata_schema}
    
I provide you with a sample, filled out metadata schema. It may contain missing information but serves as a reference to what a metadata file looks like. 
You can use it as a guide to better structure your queries. 
Sample metadata: {sample_metadata}
    
Your task is to read the user's question, which will adhere to certain guidelines or formats. 
You will only be prompted with a record ID number, and your only task is to retrieve the record and summarize information related to the modality used and subject information.
Include information about the modalities used and the subject genotype.
    
Here are some examples:
Input: <id> 719f0ac6-7d01-4586-beb9-21f52c422590 </id>
Output:
<summary> 
This record contains metadata about a behavioral experiment session with a mouse (subject ID 711039). The session involved a foraging task with auditory go cues and fiber photometry recordings.
The mouse had a Dbh-Cre genotype and was injected with a jGCaMP8m virus bilaterally in the locus coeruleus region. Optical fibers were implanted at those injection sites.
During the ~85 minute session, the mouse completed 564 trials and earned 0.558 mL of water reward through correct lick responses to auditory go cues (7.5 kHz tones at 71 dB). 
Fiber photometry data was simultaneously recorded from the four implanted fibers, with 20 μW output power per fiber. Video data was recorded from two cameras monitoring the mouse's face/body.
The behavioral rig had an enclosure, reward delivery spouts, speakers, LED light sources, filters, lenses and CMOS cameras for photometry.
</summary>

Input: </id> 2dc06357-cc30-4fd5-9e8b-f7fae7e9ba5d </id>
Output
<summary>
This record contains metadata for a behavior experiment conducted on subject 719360, a male C57BL6J mouse born on 2024-01-03. 
The experiment was performed at the Allen Institute for Neural Dynamics on 2024-04-08 using a disc-shaped mouse platform and visual/auditory stimuli presented on a monitor and speaker. 
The mouse underwent surgery for a craniotomy and headframe implantation prior to the experiment. 
During the ~1 hour session, the mouse performed a dynamic routing task with visual grating and auditory noise stimuli, consuming 0.135 mL of water reward. 
The data is stored at s3://aind-private-data-prod-o5171v/behavior_719360_2024-04-08_13-07-29 and the metadata status is listed as Invalid, though no specific issues are noted.
</summary>
    
Note: Provide the query in curly brackets, appropirately place quotation marks. 

When retrieving experiment names, pull the information through the data description module.

Even though the nature of mongodb queries is to provide false statements with the word false, in this case you will convert all words like false and null to strings -- "false" or "null".
    
When asked to provide a query, use tools, execute the query in the database, and return the retrieved information. 

Summarize the retrieved asset in natural language. 

If you are unable to provide an answer, decline to answer. Do not hallucinate an answer. Decline to answer instead.
"""


from sshtunnel import SSHTunnelForwarder
import logging, os, boto3
from urllib.parse import quote_plus
from langchain_community.vectorstores.documentdb import DocumentDBVectorSearch
from langchain_aws import BedrockEmbeddings
from pymongo import MongoClient
from motor.motor_asyncio import AsyncIOMotorClient


BEDROCK_CLIENT = boto3.client(
    service_name="bedrock-runtime",
    region_name = 'us-west-2'
)

BEDROCK_EMBEDDINGS = BedrockEmbeddings(model_id="amazon.titan-embed-text-v2:0",client=BEDROCK_CLIENT)

escaped_username = quote_plus(os.getenv("DOC_DB_USERNAME"))
escaped_password = quote_plus(os.getenv('DOC_DB_PASSWORD'))

CONNECTION_STRING = f"mongodb://{escaped_username}:{escaped_password}@localhost:27017/?directConnection=true&authMechanism=SCRAM-SHA-1&retryWrites=false"

CURATED_NAMESPACE = 'metadata_vector_index.LANGCHAIN_curated_assets'
CURATED_INDEX_NAME = "curated_embeddings_index"

CURATED_VECTORSTORE = DocumentDBVectorSearch.from_connection_string(
            connection_string=CONNECTION_STRING,
            namespace=CURATED_NAMESPACE,
            embedding=BEDROCK_EMBEDDINGS,
            index_name=CURATED_INDEX_NAME
            )

client = MongoClient(CONNECTION_STRING)
LANGCHAIN_COLLECTION = client['metadata_vector_index']['LANGCHAIN_ALL_curated_assets']

ALL_curated_namespace = 'metadata_vector_index.LANGCHAIN_ALL_curated_assets'
ALL_curated_index_name = "ALL_curated_embeddings_index"
ALL_CURATED_VECTORSTORE = DocumentDBVectorSearch.from_connection_string(
            connection_string=CONNECTION_STRING,
            namespace=ALL_curated_namespace,
            embedding=BEDROCK_EMBEDDINGS,
            index_name=ALL_curated_index_name
            )


def create_ssh_tunnel():
    """Create an SSH tunnel to the Document Database."""
    try:
        return SSHTunnelForwarder(
            ssh_address_or_host=(
                os.getenv("DOC_DB_SSH_HOST"),
                22,
            ),
            ssh_username=os.getenv("DOC_DB_SSH_USERNAME"),
            ssh_password=os.getenv("DOC_DB_SSH_PASSWORD"),
            remote_bind_address=(os.getenv("DOC_DB_HOST"), 27017),
            local_bind_address=(
                "localhost",
                27017,
            ),
        )
    except Exception as e:
        logging.error(f"Error creating SSH tunnel: {e}")

class ResourceManager:
    def __init__(self):
        self.ssh_server = None
        self.client = None
        self.async_client = None

    def __enter__(self):
        try:
            self.ssh_server = create_ssh_tunnel()
            self.ssh_server.start()
            logging.info("SSH tunnel opened")

            self.client = MongoClient(CONNECTION_STRING)
            self.async_client = AsyncIOMotorClient(CONNECTION_STRING)
            logging.info("Successfully connected to MongoDB")

            return self
        except Exception as e:
            logging.exception(e)
            self.__exit__(None, None, None)
            raise

    def __exit__(self, exc_type, exc_value, traceback):
        if self.client:
            self.client.close()
        if self.async_client:
            self.async_client.close()
        if self.ssh_server:
            self.ssh_server.stop()
        logging.info("Resources cleaned up")


import logging, sys, os, json
from typing import List, Optional
from typing_extensions import TypedDict
from langgraph.graph import END, StateGraph, START
from langgraph.checkpoint.memory import MemorySaver
from metadata_chatbot.agents.agentic_graph import datasource_router, query_retriever, query_grader, filter_generation_chain, doc_grader, rag_chain, db_rag_chain
# from metadata_chatbot.utils import ResourceManager

from metadata_chatbot.agents.docdb_retriever import DocDBRetriever
from metadata_chatbot.agents.agentic_graph import datasource_router, query_retriever, query_grader, filter_generation_chain, doc_grader, rag_chain, db_rag_chain

logging.basicConfig(filename='async_workflow.log', level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', filemode="w")

class GraphState(TypedDict):
    """
    Represents the state of our graph.

    Attributes:
        query: question asked by user
        generation: LLM generation
        documents: list of documents
    """

    query: str
    generation: str
    documents: List[str]
    filter: Optional[dict]
    #top_k: Optional[int] 

def route_question(state):
    """
    Route question to database or vectorstore
    Args:
        state (dict): The current graph state

    Returns:
        str: Next node to call
    """
    query = state["query"]

    source = datasource_router.invoke({"query": query})
    if source.datasource == "direct_database":
        logging.info("Entire database needs to be queried.")
        return "direct_database"
    elif source.datasource == "vectorstore":
        logging.info("Querying against vector embeddings...")
        return "vectorstore"

def generate_for_whole_db(state):
    """
    Filter database
    
    Args:
        state (dict): The current graph state

    Returns:
        state (dict): New key may be added to state, generation, which contains the answer for query asked
    """

    query = state['query']

    logging.info("Generating answer...")

    document_dict = dict()
    retrieved_dict = query_retriever.invoke({'query': query, 'chat_history': [], 'agent_scratchpad' : []})
    document_dict['mongodb_query'] = retrieved_dict['intermediate_steps'][0][0].tool_input['agg_pipeline']
    document_dict['retrieved_output'] = retrieved_dict['intermediate_steps'][0][1]

    documents = json.dumps(document_dict)
    return {"query": query, "documents": documents}

def filter_generator(state):
    """
    Filter database

    Args:
        state (dict): The current graph state

    Returns:
        state (dict): New key may be added to states, filter and top_k, which contains the MongoDB query that will be applied before retrieval
    """
    logging.info("Determining whether filter is required...")

    query = state["query"]

    query_grade = query_grader.invoke({"query": query}).binary_score
    logging.info(f"Database needs to be further filtered: {query_grade}")

    if query_grade == "yes":
        filter = filter_generation_chain.invoke({"query": query}).filter_query
        #top_k = filter_generation_chain.invoke({"query": query}).top_k
        logging.info(f"Database will be filtered using: {filter}")
        return {"filter": filter, "query": query}
    else:
        return {"filter": None, "top_k": None, "query": query}

def retrieve(state):
    """
    Retrieve documents

    Args:
        state (dict): The current graph state

    Returns:
        state (dict): New key added to state, documents, that contains retrieved documents
    """
    logging.info("Retrieving documents...")
    query = state["query"]
    filter = state["filter"]
    #top_k = state["top_k"]

    retriever = DocDBRetriever(k = 5)
    documents = retriever.get_relevant_documents(query = query, query_filter = filter)

    # # Retrieval
    # with ResourceManager() as RM:
    #     collection = RM.client['metadata_vector_index']['LANGCHAIN_ALL_curated_assets']
    #     retriever = DocDBRetriever(collection = collection, k = top_k)
    #     documents = retriever.get_relevant_documents(query = query, query_filter = filter)
    return {"documents": documents, "query": query}

def grade_documents(state):
    """
    Determines whether the retrieved documents are relevant to the question.

    Args:
        state (dict): The current graph state

    Returns:
        state (dict): Updates documents key with only filtered relevant documents
    """

    logging.info("Checking relevance of documents to question asked...")
    query = state["query"]
    documents = state["documents"]

    # Score each doc
    filtered_docs = []
    for doc in documents:
        score = doc_grader.invoke({"query": query, "document": doc.page_content})
        grade = score.binary_score
        logging.info(f"Retrieved document matched query: {grade}")
        if grade == "yes":
            logging.info("Document is relevant to the query")
            relevant_context = score.relevant_context
            filtered_docs.append(relevant_context)
        else:
            logging.info("Document is not relevant and will be removed")
            continue
    #doc_text = "\n\n".join(doc.page_content for doc in filtered_docs)
    return {"documents": filtered_docs, "query": query}

def generate_db(state):
    """
    Generate answer

    Args:
        state (dict): The current graph state

    Returns:
        state (dict): New key added to state, generation, that contains LLM generation
    """
    logging.info("Generating answer...")
    query = state["query"]
    documents = state["documents"]

    # RAG generation
    generation = db_rag_chain.invoke({"documents": documents, "query": query})
    return {"documents": documents, "query": query, "generation": generation, "filter": state.get("filter", None)}

def generate_vi(state):
    """
    Generate answer

    Args:
        state (dict): The current graph state

    Returns:
        state (dict): New key added to state, generation, that contains LLM generation
    """
    logging.info("Generating answer...")
    query = state["query"]
    documents = state["documents"]

    # RAG generation
    generation = rag_chain.invoke({"documents": documents, "query": query})
    return {"documents": documents, "query": query, "generation": generation, "filter": state.get("filter", None)}

workflow = StateGraph(GraphState) 
workflow.add_node("database_query", generate_for_whole_db)  
workflow.add_node("filter_generation", filter_generator)  
workflow.add_node("retrieve", retrieve)  
workflow.add_node("document_grading", grade_documents)  
workflow.add_node("generate_db", generate_db)  
workflow.add_node("generate_vi", generate_vi)  

workflow.add_conditional_edges(
    START,
    route_question,
    {
        "direct_database": "database_query",
        "vectorstore": "filter_generation",
    },
)
workflow.add_edge("database_query", "generate_db") 
workflow.add_edge("generate_db", END)
workflow.add_edge("filter_generation", "retrieve")
workflow.add_edge("retrieve", "document_grading")
workflow.add_edge("document_grading","generate_vi")
workflow.add_edge("generate_vi", END)


app = workflow.compile()

# query = "What are the injections for SmartSPIM_675387_2023-05-23_23-05-56?"

# inputs = {"query" : query}
# answer = app.invoke(inputs)
# print(answer['generation'])


from aind_data_access_api.document_db import MetadataDbClient
from aind_data_access_api.document_db_ssh import DocumentDbSSHClient, DocumentDbSSHCredentials
import json

API_GATEWAY_HOST = "api.allenneuraldynamics.org"
DATABASE = "metadata_index"
COLLECTION = "data_assets"

docdb_api_client = MetadataDbClient(
   host=API_GATEWAY_HOST,
   database=DATABASE,
   collection=COLLECTION,
)

#credentials = DocumentDbSSHCredentials()

def doc_retrieval(filter_query: dict) -> list:
    """Given a MongoDB query, this function retrieves and returns the appropriate documents.

    Parameters
    ----------
    filter_query
        MongoDB query

    Returns
    -------
    list
        List of retrieved documents
    """
    limit = 1000
    paginate_batch_size = 1000
    response = docdb_api_client.retrieve_docdb_records(
       filter_query=filter_query,
       limit=limit,
       paginate_batch_size=paginate_batch_size
    )
    return(response)

def projection_retrieval(filter_query: dict, field_name_list: list) -> list:
    """Given a MongoDB query and list of projections, this function retrieves 
    and returns the appropriate projections in the documents.

    Parameters
    ----------
    credentials 
        DocDB credentials, initialized through DocumentDbSSHCredentials

    filter_query
        MongoDB query

    field_name_list
        Field names to specifically retrieve from documents

    Returns
    -------
    list
        List of retrieved documents
    """
    projection = {"name" : 1}
    if field_name_list:
        for field_name in field_name_list:
            projection[field_name] = 1

    response = docdb_api_client.retrieve_docdb_records(
        filter_query=filter_query,
        projection=projection
    )     
    return response

def aggregation_retrieval(agg_pipeline: list) -> list:
    """Given a MongoDB query and list of projections, this function retrieves and returns the 
    relevant information in the documents. 
    Use a project stage as the first stage to minimize the size of the queries before proceeding with the remaining steps.
    The input to $map must be an array not a string, avoid using it in the $project stage.

    Parameters
    ----------
    agg_pipeline
        MongoDB aggregation pipeline

    Returns
    -------
    list
        List of retrieved documents
    """
    result = docdb_api_client.aggregate_docdb_records(
        pipeline=agg_pipeline
    )
    return result

def tool_call(tool_name:str, tool_inputs:dict) -> str:

    if tool_name == 'doc_retrieval':
        filter_query = json.loads(tool_inputs['filter'])
        retrieved_info_list = doc_retrieval(filter_query) #retrieved info type, dictionary
                    
    elif tool_name == 'projection_retrieval':
        filter_query = json.loads(tool_inputs['filter'])
        field_name_list = json.loads(tool_inputs['fieldNameList'])
        retrieved_info_list = projection_retrieval(filter_query, field_name_list)
        #retrieved_info = json.dumps(retrieved_info_list)[:1000]
                
    elif tool_name == 'aggregation_retrieval':
        #print("Loading agg pipeline...")
        agg_pipeline = json.loads(tool_inputs['pipeline'])
        #print(type(tool_inputs['pipeline']))
        retrieved_info_list = aggregation_retrieval(agg_pipeline)
        #print("Retrieved info ready")
    
    retrieved_info = " ".join(map(str, retrieved_info_list))
    #print(retrieved_info)
    return(retrieved_info)



from typing import Any, Dict, Iterator, List, Mapping, Optional

from langchain_core.callbacks.manager import CallbackManagerForLLMRun
from langchain_core.language_models.llms import LLM
from langchain_core.outputs import GenerationChunk

import boto3, json, os
from chat import get_completion
from botocore.exceptions import ClientError

bedrock = boto3.client(
    service_name="bedrock-runtime",
    region_name = 'us-west-2'
)

#model_id = "anthropic.claude-3-sonnet-20240229-v1:0"

class Metamorph(LLM):

    def _call(self, 
              prompt: str, 
              bedrock_client = bedrock, 
              stop: Optional[List[str]] = None, 
              run_manager: Optional[CallbackManagerForLLMRun] = None
              )-> str:
        answer = get_completion(prompt, bedrock_client)
        return answer
            
    
    @property
    def _identifying_params(self) -> Dict[str, Any]:
        """Return a dictionary of identifying parameters."""
        return {"model_name" : "Metamorph"}
    
    @property
    def _llm_type(self) -> str:
        """Get the type of language model used by this chat model. Used for logging purposes only."""
        return "Claude 3 Sonnet"
    
if __name__ == '__main__': 
    llm = Metamorph()
    prompt = "Give me the count of genotypes in the ecephys modality in the database."
    llm.invoke(prompt)



from langchain_community.vectorstores.documentdb import (
    DocumentDBSimilarityType,
    DocumentDBVectorSearch,
)
from langchain.docstore.document import Document 
import pymongo, os, boto3, re, json, sys
from pymongo import MongoClient
from langchain_aws import BedrockEmbeddings
import logging
from tqdm import tqdm
from bson import json_util
from langchain_text_splitters import RecursiveJsonSplitter

sys.path.append(os.path.abspath("C:/Users/sreya.kumar/Documents/GitHub/metadata-chatbot"))
from metadata_chatbot.utils import create_ssh_tunnel, CONNECTION_STRING, BEDROCK_CLIENT, ResourceManager

logging.basicConfig(filename='vector_store.log', level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', filemode="w")

TOKEN_LIMIT = 8192
JSON_SPLITTER = RecursiveJsonSplitter(max_chunk_size=TOKEN_LIMIT)

BEDROCK_EMBEDDINGS = BedrockEmbeddings(model_id = "amazon.titan-embed-text-v2:0",client = BEDROCK_CLIENT)


def regex_modality_PHYSIO(record_name: str) -> bool:

    PHYSIO_modalities = ["behavior", "Other", "FIP", "phys", "HSFP"]
    #SPIM_modalities = ["SPIM", "HCR"]

    PHYSIO_pattern = '(' + '|'.join(re.escape(word) for word in PHYSIO_modalities) + ')_'
    regex = re.compile(PHYSIO_pattern)

    return bool(regex.search(record_name))


def json_to_langchain_doc(json_doc: dict) -> tuple[list, list]:

    docs = []
    large_docs = []

    PHYSIO_fields_to_embed = ["rig", "session"]

    SPIM_fields_To_embed = ["instrument", "acquisition"]

    general_fields_to_embed = ["data_description", "subject", "procedures"]

    if regex_modality_PHYSIO(json_doc["name"]):
        fields_to_embed = [*PHYSIO_fields_to_embed, *general_fields_to_embed]
    else: 
        fields_to_embed = [*SPIM_fields_To_embed, *general_fields_to_embed]

    #fields_to_metadata = ["_id", "created", "describedBy", "external_links", "last_modified", "location", "metadata_status", "name", "processing", "schema_version"]

    to_metadata = dict()
    values_to_embed = dict()

    for item, value in json_doc.items():
        if item == "_id":
            item = "original_id"
        if item in fields_to_embed:
            values_to_embed[item] = value
        else:
            to_metadata[item] = value

    subject = json_doc.get("subject")
    
    if subject is not None:
        to_metadata["subject_id"] = subject.get("subject_id", None)  # Default if subject_id is missing
    else:
        #print("Subject key is missing or None.")
        to_metadata["subject_id"] = "null"

    data_description = json_doc.get("data_description")

    if data_description is not None:
        to_metadata["modality"] = data_description.get("modality", None)  # Default if subject_id is missing
    else:
        #print("Subject key is missing or None.")
        to_metadata["modality"] = "null"

    json_chunks = JSON_SPLITTER.split_text(json_data=values_to_embed, convert_lists=True)

    for chunk in json_chunks:
        newDoc = Document(page_content=chunk, metadata=to_metadata)
        if len(chunk) < TOKEN_LIMIT:
            docs.append(newDoc)
        else:
            large_docs.append(newDoc)

    return docs, large_docs

#INDEX_NAME = 'ALL_curated_embeddings_index'
INDEX_NAME = 'TOKEN_LIMIT_curated_embeddings_index'
NAMESPACE = 'metadata_vector_index.curated_assets'
DB_NAME, COLLECTION_NAME = NAMESPACE.split(".")


with ResourceManager() as RM:

    collection = RM.client[DB_NAME][COLLECTION_NAME]
    langchain_collection = RM.client[DB_NAME]['bigger_LANGCHAIN_curated_chunks']
    LANGCHAIN_NAMESPACE = 'metadata_vector_index.bigger_LANGCHAIN_curated_chunks'

    logging.info(f"Finding assets that are already embedded...")

    if langchain_collection is not None:
        existing_ids = set(doc['original_id'] for doc in langchain_collection.find({}, {'original_id': 1}))

    logging.info(f"Skipped {len(existing_ids)} assets, which are already in the new collection")

    docs_to_vectorize = collection.count_documents({'_id': {'$nin': list(existing_ids)}})

    logging.info(f"{docs_to_vectorize} assets need to be vectorized")

    if docs_to_vectorize != 0:

        cursor = collection.find({'_id': {'$nin': list(existing_ids)}})

        docs = []
        skipped_docs = []

        logging.info("Chunking documents...")

        document_no = 0

        for document in tqdm(cursor, desc="Chunking in progress"):
            if document_no % 100 == 0:
                logging.info(f"Currently on asset number {document_no}")
            json_doc = json.loads(json_util.dumps(document))
            chunked_docs, large_docs = json_to_langchain_doc(json_doc)
            docs.extend(chunked_docs)
            skipped_docs.extend(large_docs)
            document_no += 1

        logging.info(f"Successfully chunked {document_no} documents")

        logging.info(f"Adding {len(docs)} chunked documents to collection")
        logging.info(f"Skipping {len(skipped_docs)} due to token limitations")

        try:
            vectorstore = DocumentDBVectorSearch(
                embedding=BEDROCK_EMBEDDINGS,
                collection=langchain_collection,
                index_name=INDEX_NAME,
            )

            batch_size = 100
            for i in range(0, len(docs), batch_size):
                batch = docs[i:i + batch_size]
                vectorstore.add_documents(batch)
                logging.info(f"Added batch {i // batch_size + 1} of documents")

            dimensions = 1024
            similarity_algorithm = DocumentDBSimilarityType.COS

            logging.info("Creating vector index with chunked documents")
            vectorstore.create_index(dimensions, similarity_algorithm)

        except Exception as e:
            logging.error(f"Error processing documents: {str(e)}")          

    else:
        logging.info("Vectorstore is up to date!")
        



from urllib.parse import quote_plus
import pymongo, os, boto3, re, pickle
from pymongo import MongoClient
from langchain_community.document_loaders.mongodb import MongodbLoader
from langchain_aws import BedrockEmbeddings
from sshtunnel import SSHTunnelForwarder
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_experimental.text_splitter import SemanticChunker
from tqdm import tqdm
from transformers import AutoTokenizer
from datetime import datetime
from tqdm.contrib.logging import logging_redirect_tqdm

import logging

logging.basicConfig(filename='embeddings.log', level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

# Constants
TOKEN_LIMIT = 8192 # TODO: update the value
BATCH_SIZE = 100


# Establishing bedrock client and embedding model 
bedrock = boto3.client(
    service_name="bedrock-runtime",
    region_name = 'us-west-2'
)

embeddings_model = BedrockEmbeddings(model_id="amazon.titan-embed-text-v2:0",client=bedrock)

logging.info("Embedding model instantiated")

#Establishing tokenizer
tokenizer = AutoTokenizer.from_pretrained("bert-base-uncased")

logging.info("Tokenizer instantiated")


#TODO : CONVERT FUNCTION TO normal tokenizer
def count_tokens(id, text):
    tokens = tokenizer.encode(text, truncation=False)
    print(f"{id}:{len(tokens)} tokens")
    if len(tokens) > TOKEN_LIMIT:
        logging.info(f"{id} has {len(tokens)} tokens.  Too large.")
        return None
    return tokens

def create_ssh_tunnel():
    """Create an SSH tunnel to the Document Database."""
    try:
        return SSHTunnelForwarder(
            ssh_address_or_host=(
                os.getenv("DOC_DB_SSH_HOST"),
                22,
            ),
            ssh_username=os.getenv("DOC_DB_SSH_USERNAME"),
            ssh_password=os.getenv("DOC_DB_SSH_PASSWORD"),
            remote_bind_address=(os.getenv("DOC_DB_HOST"), 27017),
            local_bind_address=(
                "localhost",
                27017,
            ),
        )
    except Exception as e:
        logging.error(f"Error creating SSH tunnel: {e}")


def generate_embeddings_for_batch(client: MongoClient, batch: list) -> dict:
    """Generates embeddings vectors for a batch of loaded documents
    """
    logging.info("Embedding documents...")
    db = client["metadata_vector_index"]
    result_collection = db["data_assets_vectors"]
    
    skipped_ids = []
    failed_ids = []
    batch_vectors = dict()
    with logging_redirect_tqdm():
        for doc in tqdm(batch, desc="Embeddings in progress", total = len(batch)):
            
            doc_text = doc.page_content

            pattern = r"'_id':\s*'([^']+)'"
            match = re.search(pattern, doc_text)
            if match:
                id_value = match.group(1)
            else:
                # TODO: log warning
                continue

            if result_collection.count_documents({"_id": id_value, "vector_embeddings": {"$exists": True}}):
                skipped_ids.append(id_value)
                continue

            tokens = count_tokens(id_value, doc_text)
            if tokens is None:
                failed_ids.append(id_value)
                continue
            vector = embeddings_model.embed_documents([doc_text])[0]  # Embed a single document

            batch_vectors[id_value] = vector
    logging.info("Embedding finished for batch")
    logging.info(f"Succesfully embedded {len(batch_vectors)}/{len(batch)} documents.")
    logging.warning(f"Failed for {len(failed_ids)} documents: {failed_ids}")
    logging.info(f"Skipped {len(skipped_ids)} documents: {skipped_ids}")
    return batch_vectors

def write_embeddings_to_docdb_for_batch(client: MongoClient, batch_vectors: dict) -> None:
    db = client["metadata_vector_index"]
    result_collection = db["data_assets_vectors"]
    
    for id, vector in batch_vectors.items():
       logging.info(f"Adding vector embeddings for {id} to docdb")
       filter={"_id": id}
       update={
            "$set": {
                "vector_embeddings": vector
            }
        }
       result = result_collection.update_one(filter, update, upsert=False)
       logging.info(result.raw_result)
    return

database_name = "metadata_vector_index"

   # Escape username and password to handle special characters
escaped_username = quote_plus(os.getenv("DOC_DB_USERNAME"))
escaped_password = quote_plus(os.getenv('DOC_DB_PASSWORD'))

connection_string = f"mongodb://{escaped_username}:{escaped_password}@localhost:27017/?directConnection=true&authMechanism=SCRAM-SHA-1&retryWrites=false"

try:
    #print(f"Attempting to connect with: {connection_string}")

    ssh_server = create_ssh_tunnel()
    ssh_server.start()
    logging.info("SSH tunnel opened")

    client = MongoClient(connection_string)
    
    # Force a server check
    server_info = client.server_info()
    print(f"Server info: {server_info}")
    
    logging.info("Successfully connected to MongoDB")

    
    #possibly filter criteria subject/data desc etc
    loader = MongodbLoader(
        connection_string = connection_string,
        db_name = 'metadata_vector_index',
        collection_name='data_assets'
    )

    logging.info("Loading collection...")

    documents = loader.load()
    total_docs = len(documents)
    logging.info(f"Loaded {total_docs} documents..")

    #text_splitter = RecursiveCharacterTextSplitter(
    #                    chunk_size=1000,
    #                    chunk_overlap=100
    #                )

    #text_splitter = SemanticChunker(embeddings, breakpoint_threshold_type="gradient")
    #docs = text_splitter.split_documents(documents)
    #docs_text = [doc.page_content for doc in docs]


    for i in range(0, total_docs, BATCH_SIZE):
        end = i+BATCH_SIZE if i+BATCH_SIZE<total_docs else total_docs
        batch = documents[i:end]

        batch_vectors = generate_embeddings_for_batch(client=client, batch=batch)
        write_embeddings_to_docdb_for_batch(client=client, batch_vectors=batch_vectors)
        datestamp=datetime.now().strftime("%Y%m%d_%H%M%S")
        with open(f'vector_dictionary_{i}_{datestamp}.pkl', 'wb') as f:
            pickle.dump(batch_vectors, f)
        logging.info(f"Processed batch {i}")

    logging.info("Dictionary saved successfully.")

except pymongo.errors.ServerSelectionTimeoutError as e:
    print(f"Server selection timeout error: {e}")
    print(f"Current topology description: {client.topology_description}")
except Exception as e:
    logging.exception(e)
finally:
    client.close()
    ssh_server.stop()



from langchain_community.vectorstores.documentdb import DocumentDBVectorSearch
from urllib.parse import quote_plus
import pymongo, os, boto3, sys, umap
from pymongo import MongoClient
from langchain_aws import BedrockEmbeddings
import logging
import matplotlib.pyplot as plt
import umap.umap_ as umap


import numpy as np
import pandas as pd
from sklearn.manifold import TSNE
import plotly.graph_objs as go
from plotly.subplots import make_subplots
import plotly.io as pio

sys.path.append(os.path.abspath("C:/Users/sreya.kumar/Documents/GitHub/metadata-chatbot"))
from metadata_chatbot.utils import create_ssh_tunnel, ALL_CURATED_VECTORSTORE, BEDROCK_EMBEDDINGS, CONNECTION_STRING


logging.basicConfig(filename='vector_visualization.log', level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', filemode="w")

client = MongoClient(CONNECTION_STRING)
langchain_collection = client['metadata_vector_index']['LANGCHAIN_ALL_curated_assets']



ssh_server = create_ssh_tunnel()
ssh_server.start()
logging.info("SSH tunnel opened")

logging.info("Successfully connected to MongoDB")
logging.info("Initializing connection vector store")
vectorstore = ALL_CURATED_VECTORSTORE

    # query = "subject"
    # logging.info("Starting to vectorize query...")
    # query_embedding = BEDROCK_EMBEDDINGS.embed_query(query)

    # total_documents = langchain_collection.count_documents({})
    # print(f"Total documents in collection: {total_documents}")

    # # Check indexes on the collection
    # indexes = langchain_collection.index_information()
    # print(f"Indexes on collection: {indexes}")

    # result = langchain_collection.aggregate([
    #     {
    #     '$search': {
    #         'vectorSearch': {
    #             'vector': query_embedding, 
    #             'path': 'vectorContent', 
    #             'similarity': 'cosine', 
    #             'k': 22100
    #             }
    #         }
    #     },
    #     {
    #         '$limit': 22100  # Ensure the pipeline limits the results to 22100
    #     }
    # ])

logging.info("Finding vectors...")
documents = langchain_collection.find({}, {"vectorContent": 1, "_id": 0})
logging.info("Extracting vectors...") \

embeddings_list = []
for doc in documents:
    embeddings_list.append(doc["vectorContent"])
logging.info(f"Number of vectors retrieved: {len(embeddings_list)}")
embeddings_array = np.array(embeddings_list)
logging.info("Plotting...")
reducer = umap.UMAP(n_neighbors=15, min_dist=0.1, n_components=2, random_state=42)
embedding = reducer.fit_transform(embeddings_array)
# Plot the results
plt.figure(figsize=(12, 10))
plt.scatter(embedding[:, 0], embedding[:, 1], s=3, alpha=0.5)
plt.title(f'UMAP projection of {len(embeddings_list)} embeddings')
plt.xlabel('UMAP1')
plt.ylabel('UMAP2')
plt.colorbar()
plt.show()


'''
    embeddings_list = []
    modalities_list = []

    for document in result:
        embeddings_list.append(document['vectorContent'])
        modalities_list.append(document['modality'])

    #embeddings_list.insert(0,query_embedding)

    print(len(embeddings_list))
    print(len(modalities_list))

    n_components = 3 #3D 
    embeddings_list = np.array(embeddings_list) #converting to numpy array

    print(np.shape(embeddings_list))

    

    tsne = TSNE(n_components=n_components, random_state=42, perplexity=20)
    reduced_vectors = tsne.fit_transform(embeddings_list)
    print(len(reduced_vectors))
    #reduced_vectors[0:10]

        # Create a 3D scatter plot
    scatter_plot = go.Scatter3d(
        x=reduced_vectors[:, 0],
        y=reduced_vectors[:, 1],
        z=reduced_vectors[:, 2],
        mode='markers',
        marker=dict(size=5, color='grey', opacity=0.5, line=dict(color='lightgray', width=1)),
        text=[f"Point {i}" for i in range(len(reduced_vectors))]
    )

    # Highlight the first point with a different color
    highlighted_point = go.Scatter3d(
        x=[reduced_vectors[0, 0]],
        y=[reduced_vectors[0, 1]],
        z=[reduced_vectors[0, 2]],
        mode='markers',
        marker=dict(size=8, color='red', opacity=0.8, line=dict(color='lightgray', width=1)),
        text=["Question"]
        
    )

    blue_points = go.Scatter3d(
        x=reduced_vectors[1:4, 0],
        y=reduced_vectors[1:4, 1],
        z=reduced_vectors[1:4, 2],
        z1=reduced_vectors[1:4, 3],
        z2=reduced_vectors[1:4, 4],
        z3=reduced_vectors[1:4, 5],
        mode='markers',
        marker=dict(size=8, color='blue', opacity=0.8,  line=dict(color='black', width=1)),
        text=["Top 1 Data Asset","Top 2 Data Asset","Top 3 Data Asset"]
    )

    # Create the layout for the plot
    layout = go.Layout(
        scene=dict(
            xaxis=dict(title='X'),
            yaxis=dict(title='Y'),
            zaxis=dict(title='Z'),
        ),
        title=f'3D Representation after t-SNE (Perplexity=5)'
    )


    fig = make_subplots(rows=1, cols=1, specs=[[{'type': 'scatter3d'}]])

    # Add the scatter plots to the Figure
    fig.add_trace(scatter_plot)
    fig.add_trace(highlighted_point)
    fig.add_trace(blue_points)

    fig.update_layout(layout)

    pio.write_html(fig, 'interactive_plot.html')
    fig.show()
'''

client.close()
ssh_server.stop()



from pydantic import BaseModel, Field
from langchain_aws.chat_models.bedrock import ChatBedrock
from langchain import hub
import logging
from typing import Literal
from langchain_core.output_parsers import StrOutputParser
from langchain_core.tools import tool
from langchain.agents import AgentExecutor, create_tool_calling_agent
from aind_data_access_api.document_db import MetadataDbClient
from pprint import pprint

logging.basicConfig(filename='agentic_graph.log', level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', filemode="w")

GLOBAL_VARIABLE = 90

MODEL_ID = "anthropic.claude-3-sonnet-20240229-v1:0"
LLM = ChatBedrock(
    model_id= MODEL_ID,
    model_kwargs= {
        "temperature": 0
    }
)

#determining if entire database needs to be surveyed
class RouteQuery(BaseModel):
    """Route a user query to the most relevant datasource."""

    reasoning: str = Field(
        description="Give a justification for the chosen method",
    )

    datasource: Literal["vectorstore", "direct_database"] = Field(
        description="Given a user question choose to route it to the direct database or its vectorstore.",
    )

structured_llm_router = LLM.with_structured_output(RouteQuery)
router_prompt = hub.pull("eden19/query_rerouter")
datasource_router = router_prompt | structured_llm_router


# Queries that require surveying the entire database (like count based questions)
# credentials = DocumentDbSSHCredentials()
# credentials.database = "metadata_vector_index"
# credentials.collection = "curated_assets"
API_GATEWAY_HOST = "api.allenneuraldynamics-test.org"
DATABASE = "metadata_vector_index"
COLLECTION = "curated_assets"

docdb_api_client = MetadataDbClient(
   host=API_GATEWAY_HOST,
   database=DATABASE,
   collection=COLLECTION,
)

@tool
def aggregation_retrieval(agg_pipeline: list) -> list:
    """Given a MongoDB query and list of projections, this function retrieves and returns the 
    relevant information in the documents. 
    Use a project stage as the first stage to minimize the size of the queries before proceeding with the remaining steps.
    The input to $map must be an array not a string, avoid using it in the $project stage.

    Parameters
    ----------
    agg_pipeline
        MongoDB aggregation pipeline

    Returns
    -------
    list
        List of retrieved documents
    """

    result = docdb_api_client.aggregate_docdb_records(
        pipeline=agg_pipeline
    )
    return result
        
tools = [aggregation_retrieval]
db_prompt = hub.pull("eden19/entire_db_retrieval")

#llm_with_tools = LLM.bind_tools(tools)

db_surveyor_agent = create_tool_calling_agent(LLM, tools, db_prompt)
query_retriever = AgentExecutor(agent=db_surveyor_agent, tools=tools, return_intermediate_steps = True, verbose=False)
# pprint(query_retriever.invoke({'chat_history': [],"query": "What are the data asset names using mouse 675387", "agent_scratchpad":[]}))
# class retrieve_aggregation(BaseModel):
#     """List of results retrieved from mongodb database after applying pipeline generated by the model"""

#     mongo_db_query: str = Field(
#         description="Aggregation pipeline used to retrieve information from database"
#     )
#     mongo_db_results: list= Field(
#         description="List of information retrieved from the database after applying aggregation pipeline"
#     )


# aggregation_retriever = llm_with_tools.with_structured_output(retrieve_aggregation)

# aggregation_retriever_chain = db_prompt | aggregation_retriever
# print(aggregation_retriever_chain.invoke({'chat_history': [],"query": "how many documents are stored in the collection?", 'agent_scratchpad': []}))


# Processing query
class ProcessQuery(BaseModel):
    """Binary score to check whether query requires retrieval to be filtered with metadata information to achieve accurate results."""

    binary_score: str = Field(
        description="Query requires further filtering during retrieval process, 'yes' or 'no'"
    )

query_grader = LLM.with_structured_output(ProcessQuery)
query_grade_prompt = hub.pull("eden19/processquery")
query_grader = query_grade_prompt | query_grader
#print(query_grader.invoke({"query": "What is the genotype for mouse 675387?"}).binary_score)

# Generating appropriate filter
class FilterGenerator(BaseModel):
    """MongoDB filter to be applied before vector retrieval"""

    filter_query: dict = Field(description="MongoDB filter")
    #top_k: int = Field(description="Number of documents to retrieve from the database")

filter_prompt = hub.pull("eden19/filtergeneration")
filter_generator_llm = LLM.with_structured_output(FilterGenerator)

filter_generation_chain = filter_prompt | filter_generator_llm
#print(filter_generation_chain.invoke({"query": "What is the genotype for mouse 675387?"}).filter_query)

# Check if retrieved documents answer question
class RetrievalGrader(BaseModel):
    """Binary score to check whether retrieved documents are relevant to the question"""

    reasoning: str = Field(
        description="Give a reasoning as to what makes the document relevant for the chosen method",
    )

    binary_score: str = Field(
        description="Retrieved documents are relevant to the query, 'yes' or 'no'"
    )

    relevant_context: str = Field(
        description="Relevant pieces of context in document"
    )

retrieval_grader = LLM.with_structured_output(RetrievalGrader)
retrieval_grade_prompt = hub.pull("eden19/retrievalgrader")
doc_grader = retrieval_grade_prompt | retrieval_grader
# doc_grade = doc_grader.invoke({"query": question, "document": doc}).binary_score
# logging.info(f"Retrieved document matched query: {doc_grade}")

# Generating response to documents
answer_generation_prompt = hub.pull("eden19/answergeneration")
rag_chain = answer_generation_prompt | LLM | StrOutputParser()

db_answer_generation_prompt = hub.pull("eden19/db_answergeneration")
db_rag_chain = answer_generation_prompt | LLM | StrOutputParser()
# generation = rag_chain.invoke({"documents": doc, "query": question})
# logging.info(f"Final answer: {generation}")


import sys, os, json
from typing import List, Optional, Any, Union, Annotated
from pymongo.collection import Collection
from motor.motor_asyncio import AsyncIOMotorCollection
from langchain_core.callbacks import CallbackManagerForRetrieverRun
from langchain_core.documents import Document
from langchain_core.retrievers import BaseRetriever
from bson import json_util
from langsmith import trace as langsmith_trace
from pydantic import Field
from aind_data_access_api.document_db import MetadataDbClient

sys.path.append(os.path.abspath("C:/Users/sreya.kumar/Documents/GitHub/metadata-chatbot"))
from metadata_chatbot.utils import BEDROCK_EMBEDDINGS

API_GATEWAY_HOST = "api.allenneuraldynamics-test.org"
DATABASE = "metadata_vector_index"
COLLECTION = "bigger_LANGCHAIN_curated_chunks"

docdb_api_client = MetadataDbClient(
   host=API_GATEWAY_HOST,
   database=DATABASE,
   collection=COLLECTION,
)


class DocDBRetriever(BaseRetriever):
    """A retriever that contains the top k documents, retrieved from the DocDB index, aligned with the user's query."""
    #collection: Any = Field(description="DocDB collection to retrieve from")
    k: int = Field(default=5, description="Number of documents to retrieve")

    def _get_relevant_documents(
        self, 
        query: str, 
        query_filter: Optional[dict] = None,
        run_manager: Optional[CallbackManagerForRetrieverRun] = None,
        **kwargs: Any,
    ) -> List[Document]:
        
        #Embed query
        embedded_query = BEDROCK_EMBEDDINGS.embed_query(query)

        #Construct aggregation pipeline
        vector_search = {
            "$search": { 
                "vectorSearch": { 
                    "vector": embedded_query, 
                    "path": 'vectorContent', 
                    "similarity": 'euclidean', 
                    "k": self.k
                }
            }
        }

        pipeline = [vector_search]
        if query_filter:
            pipeline.insert(0, query_filter)
    
        result = docdb_api_client.aggregate_docdb_records(pipeline=pipeline)

        page_content_field = 'textContent'

        results = []
        
        #Transform retrieved docs to langchain Documents
        for document in result:
            values_to_metadata = dict()

            json_doc = json.loads(json_util.dumps(document))

            for key, value in json_doc.items():
                if key == page_content_field:
                    page_content = value
                else:
                    values_to_metadata[key] = value

            new_doc = Document(page_content=page_content, metadata=values_to_metadata)
            results.append(new_doc)

        return results
    
    async def _aget_relevant_documents(
        self, 
        query: str, 
        query_filter: Optional[dict] = None,
        run_manager: Optional[CallbackManagerForRetrieverRun] = None,
        **kwargs: Any,
    ) -> List[Document]:
        
        #Embed query
        embedded_query = BEDROCK_EMBEDDINGS.embed_query(query)

        #Construct aggregation pipeline
        vector_search = {
            "$search": { 
                "vectorSearch": { 
                    "vector": embedded_query, 
                    "path": 'vectorContent', 
                    "similarity": 'euclidean', 
                    "k": self.k,
                    "efSearch": 40
                }
            }
        }

        pipeline = [vector_search]
        if query_filter:
            pipeline.insert(0, query_filter)
    
        result = docdb_api_client.aggregate_docdb_records(pipeline=pipeline)
        #results = await cursor.to_list(length=1000)

        page_content_field = 'textContent'

        results = []
        
        #Transform retrieved docs to langchain Documents
        for document in result:
            values_to_metadata = dict()

            json_doc = json.loads(json_util.dumps(document))

            for key, value in json_doc.items():
                if key == page_content_field:
                    page_content = value
                else:
                    values_to_metadata[key] = value

            new_doc = Document(page_content=page_content, metadata=values_to_metadata)
            results.append(new_doc)

        return results


import boto3, json, os, logging
from tools import doc_retrieval, projection_retrieval, aggregation_retrieval, tool_call
from system_prompt import system_prompt, summary_system_prompt
from config import toolConfig
from botocore.exceptions import ClientError

logging.basicConfig(filename='error.log', level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

#Connecting to bedrock

bedrock = boto3.client(
    service_name="bedrock-runtime",
    region_name = 'us-west-2'
)

model_id = "anthropic.claude-3-sonnet-20240229-v1:0"

def get_completion(prompt: str, bedrock_client, system_prompt=system_prompt) -> str:
    
    """Given a prompt, this function returns a reply to the question.

    Parameters
    ----------
    prompt: str 
        Query given by the user to the chatbot
    bedrock_client: variable
        Initialization of boto3 bedrock client
    system_prompt: str 
        Commands to be given to the model, will determine model response
    prefill: str 
        Formatted prefill words to start Claude's reply

    Returns
    -------
    str
        Model's reply to prompt
    """

    messages = [{"role": "user", "content": [{"text": prompt}]}]
    
    inference_config = {
        "temperature": 0,
        "maxTokens": 4096
    }
    converse_api_params = {
        "modelId": model_id,
        "messages" : messages,
        "inferenceConfig": inference_config,
        "toolConfig": toolConfig,
        "system" : [{"text": system_prompt}] 
    }

    try:
        response = bedrock_client.converse(**converse_api_params)  
        print(response)
        response_content_blocks = response['output']['message']['content']

        #Printing Claude's initial response to query
        print(response_content_blocks[0]['text'])
        
        #Assistant reply including tool use 
        messages.append({"role": "assistant", "content": response_content_blocks})

        if response['stopReason'] == "tool_use":
            tool_use = response_content_blocks[-1]['toolUse']
            tool_id = tool_use['toolUseId']
            tool_name = tool_use['name']
            tool_inputs = tool_use['input']
            
            logging.info(f"Using tool {tool_name}")
            
            retrieved_info = tool_call(tool_name, tool_inputs)

            tool_response = {
                            "role": "user",
                            "content": [
                                        {
                                        "toolResult": {
                                            "toolUseId": tool_id,
                                            "content": [
                                                {
                                                "text": retrieved_info
                                                }
                                            ],
                                        'status':'success'
                                        }
                                    }
                                    ]
                            }
                
            messages.append(tool_response)
            logging.info("Successful information retrieval")
                
            converse_api_params = {
                                    "modelId": model_id,
                                    "messages": messages,
                                    "inferenceConfig": inference_config,
                                    "toolConfig": toolConfig 
                                    }
            
            logging.info("Generating response...")
            final_response = bedrock_client.converse(**converse_api_params) 
            final_response_text = final_response['output']['message']['content'][0]['text']
            return(final_response_text)
                    
    except ClientError as e:
        logging.error("A client exception occurred: %s", str(e), exc_info=True)


def get_summary(prompt, bedrock_client = bedrock, system_prompt=summary_system_prompt):

    messages = [{"role": "user", "content": [{"text": f"Summarize the record with id {prompt}"}]}]
    
    inference_config = {
        "temperature": 0,
        "maxTokens": 2000
    }
    converse_api_params = {
        "modelId": model_id,
        "messages" : messages,
        "inferenceConfig": inference_config,
        "toolConfig": toolConfig
    }
    
    if system_prompt:
        converse_api_params["system"] = [{"text": system_prompt}]

    try:
        response = bedrock_client.converse(**converse_api_params)
        
        response_message = response['output']['message']
        
        response_content_blocks = response_message['content']
        
        messages.append({"role": "assistant", "content": response_content_blocks})
        
        for content_block in response_content_blocks:
            if 'toolUse' in content_block:
                
                tool_use = response_content_blocks[-1]
                tool_id = tool_use['toolUse']['toolUseId']
                tool_name = tool_use['toolUse']['name']
                tool_inputs = tool_use['toolUse']['input']
                
                if tool_name == 'doc_retrieval':
                    filter_query_s = tool_inputs['filter'] # filter query stored as a string instead of dictionary
                    filter_query = json.loads(filter_query_s)
                    retrieved_info_list = doc_retrieval(filter_query) #retrieved info type, dictionary
                    retrieved_info = " ".join(map(str, retrieved_info_list))
             
                tool_response = {
                                "role": "user",
                                "content": [
                                            {
                                            "toolResult": {
                                                "toolUseId": tool_id,
                                                "content": [
                                                    {
                                                    "text": retrieved_info
                                                    }
                                                ],
                                            'status':'success'
                                            }
                                        }
                                        ]
                                }
                    
                messages.append(tool_response)
                    
                converse_api_params = {
                                        "modelId": model_id,
                                        "messages": messages,
                                        "inferenceConfig": inference_config,
                                        "toolConfig": toolConfig 
                                        }

                final_response = bedrock_client.converse(**converse_api_params) 
                final_response_text = final_response['output']['message']['content'][0]['text']
                return(final_response_text)
                    
    except ClientError as err:
        message = err.response['Error']['Message']
        print(f"A client error occured: {message}")
        
        
def simple_chat(bedrock_client = bedrock, system_prompt = system_prompt):
    
    """This function is able to demonstrate back and forth conversation given user input.

    Parameters
    ----------
    bedrock_client: variable
        Initialization of boto3 bedrock client
    system_prompt: str 
        Commands to be given to the model, will determine model response

    Returns
    -------
    str
        Model's reply to prompt
    """

    model_id = "anthropic.claude-3-sonnet-20240229-v1:0"
    
    user_message = input("\nUser: ")
    messages = [{"role": "user", "content": [{"text": user_message}]}]
    
    inference_config = {
        "temperature": 0,
        "maxTokens": 4000
    }
    
    while True:
        #If the last message is from the assistant, get another input from the user
        if messages[-1].get("role") == "assistant":
            user_message = input("\nUser: ")
            messages.append({"role": "user", "content": [{"text": user_message}]})

        converse_api_params = {
            "modelId": model_id,
            "messages": messages,
            "inferenceConfig": inference_config,
            "toolConfig":toolConfig,
        }
        if system_prompt:
            converse_api_params["system"] = [{"text": system_prompt}]

        response = bedrock_client.converse(**converse_api_params)
        print(response)

        messages.append({"role": "assistant", "content": response['output']['message']['content']})

        #If Claude stops because it wants to use a tool:
        if response['stopReason'] == "tool_use":
            tool_use = response['output']['message']['content'][-1] #Naive approach assumes only 1 tool is called at a time
            tool_id = tool_use['toolUse']['toolUseId']
            tool_name = tool_use['toolUse']['name']
            tool_inputs = tool_use['toolUse']['input']

            print(f"Using the {tool_name} tool...")
            print(f"Tool Input:")
            print(json.dumps(tool_inputs, indent=2))
            
            retrieved_info = tool_call(tool_name, tool_inputs)
            

            messages.append({
                "role": "user",
                "content": [
                    {
                        "toolResult": {
                            "toolUseId": tool_id,
                                    "content": [
                                            {
                                                "text": retrieved_info
                                             }
                                                    ],
                        
                        }
                    }
                ]
            })

        else: 
            print("\nClaude:" + f"{response['output']['message']['content'][0]['text']}")
            #print("\nClaude: Is there anything else I can help you with?")

if __name__ == '__main__': 
    #simple_chat(bedrock)
    prompt = "What is the experimental history for subject 664956"
    response = get_completion(prompt, bedrock)
    print(response)


from typing import Any, Dict, Iterator, List, Mapping, Optional

from langchain_core.callbacks.manager import CallbackManagerForLLMRun
from langchain_core.language_models.llms import LLM
from langchain_core.outputs import GenerationChunk

import logging, asyncio

from metadata_chatbot.agents.async_workflow import async_app
from metadata_chatbot.agents.workflow import app


class GAMER(LLM):

    def _call(
        self,
        query: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> str:
        """
        Args:
            query: Natural language query.
            stop: Stop words to use when generating. Model output is cut off at the
                first occurrence of any of the stop substrings.
                If stop tokens are not supported consider raising NotImplementedError.
            run_manager: Callback manager for the run.
            **kwargs: Arbitrary additional keyword arguments. These are usually passed
                to the model provider API call.

        Returns:
            The model output as a string.
        """
        inputs = {"query" : query}
        answer = app.invoke(inputs)
        return answer['generation']

        # inputs = {"query" : query}                          
        # answer = app.invoke(inputs)                                           
        # if isinstance(answer, dict):
        #     result_text = answer.get('text', '')  # Adjust as necessary based on the actual structure
        # else:
        #     result_text = answer
        # print(result_text)
        # return answer
    
    async def _acall(
        self,
        query: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> str:
        """
        Asynchronous call.

        Args:
            query: Natural language query.
            stop: Stop words to use when generating. Model output is cut off at the
                first occurrence of any of the stop substrings.
                If stop tokens are not supported consider raising NotImplementedError.
            run_manager: Callback manager for the run.

        Returns:
            The model output as a string.
        """
        inputs = {"query" : query}
        async for output in async_app.astream(inputs, stream_mode="updates"):
            for key, value in output.items():
                logging.info(f"Currently on node '{key}':")
        return value['generation'] if value else None

    def _stream(
        self,
        query: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[GenerationChunk]:
        """Stream the LLM on the given prompt.

        This method should be overridden by subclasses that support streaming.

        If not implemented, the default behavior of calls to stream will be to
        fallback to the non-streaming version of the model and return
        the output as a single chunk.

        Args:
            query: The prompt to generate from.
            stop: Stop words to use when generating. Model output is cut off at the
                first occurrence of any of these substrings.
            run_manager: Callback manager for the run.
            **kwargs: Arbitrary additional keyword arguments. These are usually passed
                to the model provider API call.

        Returns:
            An iterator of GenerationChunks.
        """
        for char in query[: self.n]:
            chunk = GenerationChunk(text=char)
            if run_manager:
                run_manager.on_llm_new_token(chunk.text, chunk=chunk)

            yield chunk

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        """Return a dictionary of identifying parameters."""
        return {
            "model_name": "Anthropic Claude 3 Sonnet",
        }

    @property
    def _llm_type(self) -> str:
        """Get the type of language model used by this chat model. Used for logging purposes only."""
        return "Claude 3 Sonnet"
    
# llm = GAMER()
# print(llm.invoke("What are the injections for SmartSPIM_675387_2023-05-23_23-05-56?"))

# async def main():
#     result = await llm.ainvoke("Can you give me a timeline of events for subject 675387?")
#     print(result)

# asyncio.run(main())


from fastapi import FastAPI
import uvicorn
from metadata_chatbot.bedrock_model.chat import get_summary

app = FastAPI()

@app.get("/summary/{_id}")
def REST_summary(_id: str):
    result = get_summary(_id)
    return result

if __name__ == "__main__":
    uvicorn.run(app, host="127.0.0.1", port=8000)